import numpy as np
import sys
sys.path.insert(1, '/homes_unix/agillig/Projects/DynaPred/code')

import edge_ts as ets

def framewise_selection(X, lower_q, upper_q, criterion='RSS'):
    """
    Select frames based on the given criterion.
    
    Parameters
    ----------
    X : array-like, rs timeseries; shape (n_features, n_TR)
        The input data.

    lower_q : float
        The lower quantile.
    upper_q : float
        The upper quantile.
    criterion : str, default='RSS'
        The criterion to use for selecting frames.
        Possible values are 'RSS' and 'BIC'.
    
    Returns
    -------
    X_selected : array-like, shape (n_edges, n_selected_TR)
    """
    n_edges, n_TR = X.shape

    edge_ts = ets.compute_edge_ts(X)

    if criterion == 'RSS':
        rss = np.sqrt(np.sum(edge_ts**2, axis=1)).reshape(1,-1)
        selector = rss
        # print(f'shape of RSS: {rss.shape}')
    else:
        raise ValueError("Invalid criterion: %s, valid choices are: RSS (yes that's it)" % criterion)
    #identify the lower and upper quantile of the RSS
    lower = np.percentile(selector, lower_q)
    upper = np.percentile(selector, upper_q)

    lower_mask = selector > lower
    upper_mask = selector < upper
    selection_mask = np.logical_and(lower_mask, upper_mask).squeeze()
    X_selected = X[:, selection_mask]

    return X_selectedimport numpy as np
import sys
sys.path.insert(1, '/homes_unix/agillig/Projects/DynaPred/code')

import edge_ts as ets

def framewise_selection(X, lower_q, upper_q, criterion='RSS'):
    """
    Select frames based on the given criterion.
    
    Parameters
    ----------
    X : array-like, rs timeseries; shape (n_features, n_TR)
        The input data.

    lower_q : float
        The lower quantile.
    upper_q : float
        The upper quantile.
    criterion : str, default='RSS'
        The criterion to use for selecting frames.
        Possible values are 'RSS' and 'BIC'.
    
    Returns
    -------
    X_selected : array-like, shape (n_edges, n_selected_TR)
    """
    n_edges, n_TR = X.shape

    edge_ts = ets.compute_edge_ts(X)

    if criterion == 'RSS':
        rss = np.sqrt(np.sum(edge_ts**2, axis=1)).reshape(1,-1)
        selector = rss
        # print(f'shape of RSS: {rss.shape}')
    else:
        raise ValueError("Invalid criterion: %s, valid choices are RSS (yes that's it)" % criterion)
    #identify the lower and upper quantile of the RSS
    lower = np.percentile(selector, lower_q)
    upper = np.percentile(selector, upper_q)

    lower_mask = selector > lower
    upper_mask = selector < upper
    selection_mask = np.logical_and(lower_mask, upper_mask).squeeze()
    X_selected = X[:, selection_mask]

    return X_selected



import sqlite3
import pandas as pd
import numpy as np
import parcellate as par
from pathlib import Path
import os
from abc import ABC
import warnings
from tqdm.auto import tqdm
from concurrent.futures import ThreadPoolExecutor
import time
import functional_connectivity as FC
from joblib import Parallel, delayed
import warnings
from threading import Lock

def RetrieveTablesFromSQL(db_file, table_names=None, index_col=None, n_jobs=1):
    #TODO: implement joblib / mutltithread parallelization
    conn = sqlite3.connect(db_file)
    cursor = conn.cursor()

    cursor.execute("SELECT name FROM sqlite_master WHERE type='table'")
    tables = cursor.fetchall()
    tables = [table[0] for table in tables]
    cursor.close()
    
    if table_names is not None:
        tables = table_names
        if type(tables) == str:
            query = f"SELECT * FROM {tables}"
            data = pd.read_sql_query(query, conn, index_col=index_col)
            return data, tables
    conn.close()
    
    data = []
    # for table in tqdm(tables):
        # query = f"SELECT * FROM {table}"
        # data.append(pd.read_sql_query(query, conn, index_col=index_col))
        # conn.close()
    # args = [(db_file, table, index_col) for table in tables]
    timeout=9999
    # Ignore UserWarning: A worker stopped while some jobs were given to the executor
    warnings.filterwarnings("ignore", message="A worker stopped while some jobs were given to the executor")
    data = Parallel(n_jobs=n_jobs, timeout=timeout)(delayed(retrieve_table)(db_file, table_name, index_col) for table_name in tqdm(tables))
    return data, tables
    

    df_list = []
    for table_name in tqdm(tables):
        if type(table_name) == 'tuple':
            table_name = table_name[0]
        query = f"SELECT * FROM {table_name}"
        df = pd.read_sql_query(query, conn, index_col=index_col)
        df_list.append(df)

    return df_list

def retrieve_table(db_file, table_name, index_col, conn=None):
    with sqlite3.connect(db_file) as conn:
        query = f"SELECT * FROM {table_name}"
        return pd.read_sql_query(query, conn, index_col=index_col)


def SaveToSQLTable(data, db_file, index_label=None, table_name='data', **kwargs):
    # if isinstance(data, pd.DataFrame) == False:
    #     if data.shape[0] != 1:
    # data = data.reshape(1, -1)
    
    if index_label is None:
        index = False
    else:
        index = True
    conn = sqlite3.connect(db_file)
    lock = Lock()
    with lock:
        while True:
            try:
                data.to_sql(f'{table_name}',
                                conn,
                                if_exists='append',
                                index=index,
                                index_label=index_label)
                conn.close()
                break
            except sqlite3.OperationalError:
                time.sleep(0.01)
                data.to_sql(f'{table_name}',
                                conn,
                                if_exists='append',
                                index=index,
                                index_label=index_label)
    conn.close()


def save_to_sql_in_memory(data, conn, index_label=None, table_name='data', **kwargs):
    # if isinstance(data, pd.DataFrame) == False:
    #     if data.shape[0] != 1:
    # data = data.reshape(1, -1)
    data = pd.DataFrame(data)
    
    if index_label is None:
        index = False
    else:
        index = True

    # lock = Lock()

    # with lock:
    while True:
        try:
            data.to_sql(f'{table_name}',
                            conn,
                            if_exists='append',
                            index=index,
                            index_label=index_label)
            break
        except sqlite3.OperationalError:
            time.sleep(0.01)
            data.to_sql(f'{table_name}',
                            conn,
                            if_exists='append',
                            index=index,
                            index_label=index_label)



def get_data(SubID, database):
    #database: str referring to the fMRI database. Can be "ISHARE" or "MEMENTO".
    
    exists = check_data_existence(SubID, database)

    if exists:
        return retrieve_from_database(SubID, database)
    else:
        print(f'{SubID} not in the database, computing and saving parcellation')
        # return parcellate_and_retrieve(SubID, database)
        pass

def get_fc_data(db_file, SubList=None, check_missing=True, compute_missing=True, parcellation_db=None, matrix_form=False, atlas=None, **kwargs):
    args = (db_file, SubList, check_missing, compute_missing,parcellation_db, matrix_form, atlas)
    #TODO: add support for parcellated data as pandas df / np array
    #TODO: make it abstract and applicable to different fetching types such as pickle, csv, etc.
    #TODO: take out automatic computing of missing subjects. this should be handled by another function
    if type(SubList) == 'str':
        SubList = [SubList]

    conn = sqlite3.connect(db_file)
    cur = conn.cursor()
    # Execute a SQL query to get all table names
    cur.execute("SELECT name FROM sqlite_master WHERE type='table'")
    # Fetch all results
    table_names = cur.fetchall()
    # Close the cursor and connection
    cur.close()
    conn.close()
    table_names = [table[0] for table in table_names]
    missing = check_missing_datatables(SubList, table_names)

    #TODO: ignore files for which parcellation is also missing

    if len(missing) == 0 or check_missing == False:
        return retrieve_from_sql_table(db_file, index_col=None)
    else:
        raise ValueError(f'{len(missing)} subjects not in the database, compute and save functional connectivity first')
        # return parcellate_and_retrieve(SubID, database)
        if parcellation_db is None:
            raise ValueError('parcellation_db is required to compute functional connectivity')
        # for SubID in tqdm(missing):
        #     if check_table_in_sql(SubID, parcellation_db) == False or compute_missing == False:
        #         continue
        #     data = retrieve_parcellated_data(SubID, parcellation_db, atlas=atlas, compute_missing=compute_missing)
        #     FC.compute_connectivity(data, matrix_form=matrix_form)
        #     SaveToSQLTable(SubID, data, db_file, SubID)

        Parallel(n_jobs=40)(delayed(process_fc_subid)(SubID, parcellation_db, atlas, compute_missing, matrix_form, db_file) for SubID in tqdm(missing))
        return get_fc_data(*args, **kwargs)

def process_fc_subid(SubID, parcellation_db, atlas, compute_missing, matrix_form, db_file):
    if check_table_in_sql(SubID, parcellation_db) == False or compute_missing == False:
        return
    data = retrieve_parcellated_data(SubID, parcellation_db, atlas=atlas, compute_missing=compute_missing)
    fc_data = FC.compute_connectivity(data, matrix_form=matrix_form)
    max_retry = 20
    c = 0
    while c < max_retry:
        try:
            return SaveToSQLTable(fc_data, db_file, SubID)
        except sqlite3.OperationalError:
            time.sleep(0.01)
            c += 1
            continue


def check_missing_datatables(SubList, table_names):
    return [SubID for SubID in SubList if SubID not in table_names]


def check_data_in_sql_table(SubID, db_file, colname='SHARE_ID'):
    #works
    conn = sqlite3.connect(db_file)

    cur = conn.cursor()
    # Execute a SQL query to get all table names
    cur.execute("SELECT name FROM sqlite_master WHERE type='table'")
    # Fetch all results
    table_names = cur.fetchall()
    table_name = table_names[0][0]
    # Close the cursor and connection
    cur.close()

    cur = conn.cursor()
    # Execute a SQL query to check if the table exists
    cur.execute(f"SELECT EXISTS(SELECT 1 FROM {table_name} WHERE {colname} = '{SubID}')")

    # Fetch the result
    in_df = cur.fetchone()[0]
    # Close the cursor
    cur.close()
    conn.close()
    # Check if the table exists
    if in_df:
        return True
    else:
        return False
    
def check_table_in_sql(table_name, db_file):
    conn = sqlite3.connect(db_file)
    cur = conn.cursor()
    # Execute a SQL query to check if the table exists
    cur.execute(f"SELECT name FROM sqlite_master WHERE type='table' AND name='{table_name}';")

    # Fetch the result
    in_df = cur.fetchone()

    if in_df is not None:
        in_df = in_df[0]   

    # Close the cursor
    cur.close()

    # Check if the table exists
    if in_df:
        return True
    else:
        return False
    
def check_tables_in_sql(SubList, db_file):
    conn = sqlite3.connect(db_file)
    cur = conn.cursor()
    # Execute a SQL query to check if the table exists
    cur.execute(f"SELECT name FROM sqlite_master WHERE type='table';")

    # Fetch the result
    in_df = cur.fetchall()
    in_df = [table[0] for table in in_df]
    cur.close()
    conn.close()

    missing = [SubID for SubID in SubList if SubID not in in_df]
    return missing

    

def check_missing_tables(SubList, db_file):
    #terribly inefficient
    return [SubID for SubID in SubList if not check_table_in_sql(SubID, db_file)]
class DataRetriever(ABC):
    # @abstractmethod
    def retrieve(self, SubID):
        pass

class ISHARERetriever(DataRetriever):
    def retrieve(self, SubID):
        pass


def retrieve_from_database(SubID, database):
    pass
    # if database == 'ISHARE':
    #     retriever = 

def retrieve_parcellated_data(SubID, db_file, DataFile=None, compute_missing=True, **kwargs):
    """
    Retrieve parcellated data from the database
    """
    atlas = kwargs.get('atlas', None)

    # get the data from the database
    query = f"SELECT * FROM {SubID}"

    # Execute a SQL query and load the result into a DataFrame
    conn = sqlite3.connect(db_file)
    #check if table for subject exists
    cur = conn.cursor()

    # Execute a SQL query to check if the table exists
    cur.execute(f"SELECT name FROM sqlite_master WHERE type='table' AND name='{SubID}';")

    # Fetch the result
    in_df = cur.fetchone()

    # Close the cursor
    cur.close()

    # Check if the table exists
    if in_df:
        data = pd.read_sql_query(query, conn) 
        return data

    print(f'{SubID} not in the database, computing and saving parcellation')

    if compute_missing == False:
        print('computing missing data is disabled, exciting')
        return

    # if the data is not in the database, parcellate the data
    if atlas is None:
        raise ValueError(f'an atlas is required to parcellate the data. subject {SubID} not found in the database {db_file} and no atlas provided')
    #atlas = Path(".").absolute().parent / "atlases/"
    if DataFile is None:
        raise ValueError('DataFile is required to parcellate the data')
    img = ''.join(DataFile)

    data = par.parcellate(img, atlas)
    data = pd.DataFrame(data)
    try:
        data.to_sql(f'{SubID}',
                    conn,
                    index=True,
                    index_label='ROI')
    except sqlite3.OperationalError:
        time.sleep(0.01)
        data.to_sql(f'{SubID}',
                    conn,
                    index=True,
                    index_label='ROI')

        
    
    
    conn.commit()
    conn.close()

    return data

def retrieve_all_parcellated_data(db_file, SubList, **kwargs):
    #TODO, add condition to check n_subjects vs len(database); compute the missing subjects
    compute_missing = kwargs.get('compute_missing', False)
    conn = sqlite3.connect(db_file)
    cursor = conn.cursor()
    # Get all table names
    cursor.execute("SELECT name FROM sqlite_master WHERE type='table'")
    tables = cursor.fetchall()
    tables = [table[0] for table in tables]
    # Retrieve all tables and store them in a list
    df_list = []

    missing = check_missing_datatables(SubList, tables)

    if len(missing) != 0:
        print('missing parcellations for some subjects')
        if compute_missing:
            parcellation_db = kwargs.get('parcellation_db', None)
            atlas = kwargs.get('atlas', None)
            if parcellation_db is None or atlas is None:
                raise ValueError('parcellation_db and atlas are required to compute missing subjects')
            for SubID in list(missing):
                data = retrieve_parcellated_data(SubID, parcellation_db, atlas=atlas)
                retry(SaveToSQLTable, SubID, data, db_file, SubID)
        else:
            print('missing parcellations will not be computed')


    with ThreadPoolExecutor() as executor:
        for table_name in tqdm(tables):
            query = f"SELECT * FROM {table_name}"
            # df = pd.read_sql_query(query, conn)
            df = pd.read_sql_query(query,
                                   conn
                                   )
            df_list.append(df)

    # df = pd.concat(df_list, axis=2)

    return df_list

def retrieve_from_sql_table(db_file, SubID=None, index_col="SHARE_ID"):
    #NOT TESTED; TEMPLATE
    conn = sqlite3.connect(db_file)
    cur = conn.cursor()
    # Execute a SQL query to get all table names
    cur.execute("SELECT name FROM sqlite_master WHERE type='table'")
    # Fetch all results
    table_name = cur.fetchone()[0]

    # Get the table info
    cur.execute(f"PRAGMA table_info({table_name})")
    table_info = cur.fetchall()

    # Close the cursor and connection
    cur.close()
    
    if SubID is None:
        query = f"SELECT * FROM {table_name}"
    else:
        query = f"SELECT * FROM {table_name} WHERE {index_col} = '{SubID}'"

    data = pd.read_sql_query(query, conn, index_col=index_col)
    conn.close()
    return data

    

    


# def retry(func, *args, **kwargs):
#     max_retry_count = 10  # Maximum number of retries
#     retry_count = 0

#     while retry_count < max_retry_count:
#         try:
#             return func(*args, **kwargs)
#         except sqlite3.OperationalError as e:
#             if 'database is locked' in str(e):
#                 time.sleep(1)  # Wait for 1 second
#                 retry_count += 1
#                 if retry_count == max_retry_count:  # If this was the last retry
#                     raise Exception("Maximum number of retries exceeded")  # Raise a custom exception


import numpy as np
from scipy.spatial.distance import squareform
from scipy.stats import zscore
from itertools import product
import pandas as pd

def compute_edge_timeseries(rsData, matrix_form=False):
    
    # zscore
    rsData = zscore(rsData, axis=0)

    pairwise_products = np.array([np.outer(rsData[:, i], rsData[:, i]) for i in range(rsData.shape[1])])

    if matrix_form:
        edge_ts = pairwise_products
        # np.fill_diagonal(edge_ts, 0)
    else:
        indices_uppertriangle = [np.triu_indices(pairwise_products[0].shape[0], 1) for i in range(pairwise_products.shape[0])]
        edge_ts = np.array([prod[ind] for prod, ind in zip(pairwise_products, indices_uppertriangle)])
    return edge_ts

def label_ets(labels):
    """
    labels: name of ROIs, size n_ROIs
    return: ets_labels
    """
    indices = np.triu_indices(len(labels), 1)

    ets_labels = []
    for i, (x,y) in enumerate(zip(*indices)):
        ets_labels.append((labels[x], labels[y]))

    return ets_labels

def ets_labels_to_int(ets_labels):
    # Normalize each label by sorting the tuples
    normalized_labels = [tuple(sorted(label)) for label in ets_labels]
    
    # Create a list of unique labels using a set to remove duplicates, then convert it back to a list and sort it
    unique_labels = sorted(list(set(normalized_labels)))
    
    # Create an array of integer labels where each label corresponds to the index of the tuple in unique_labels plus one
    int_labels = np.array([unique_labels.index(label) + 1 for label in normalized_labels], dtype=int)
    
    return int_labels
        



from sklearn.base import BaseEstimator, RegressorMixin
import sys
sys.path.insert(1, '/gpfs/home/agillig/Projects/DynaPred/code/l0learn')
import l0learn

class L0Regression(BaseEstimator, RegressorMixin):

    def __init__(self):  
        self.lib_loc = '/gpfs/home/agillig/R/x86_64-pc-linux-gnu-library/4.3'
        self.n_folds = 10

    def fit(self, X, y = None):
        self.coef, self.support_ = l0learn.cvfit(X, y, n_cv=self.n_folds, lib_loc=self.lib_loc)
        return self

    def predict(self, X, y = None):
        return l0learn.predict(X, self.coef)


# class AbessRegression(BaseEstimator, RegressorMixin):

#     def __init__(self):
#         self.n_folds = 10

class ModelObject:
    def __init__(self, model, grid):
        self.model = model
        self.param_grid = grid

    def get_model(self):
        return self.model
    
    def get_param_grid(self):
        return self.param_grid


from functools import reduce
import sys
from itertools import filterfalse, groupby, combinations, chain, product
from operator import add

def unique(iterable):
    """
    List unique elements, preserving order. Remember all elements ever seen.
    Siplified version of unique_everseen from itertools recipes.
    """
    seen = set()
    for element in filterfalse(seen.__contains__, iterable):
        seen.add(element)
        yield element

def itersize(iterable):
    return reduce(
        add,
        map(sys.getsizeof, iterable)
    )

def all_equal(iterable):
    """
    From itertools recipes
    """
    g = groupby(iterable)
    return next(g, True) and not next(g, False)

def loops(it):
    unique_el = unique(it)
    return ((el, el) for el in unique_el)

def all_connectivities(it):
    return chain(combinations(it, 2), loops(it))

def join(it):
    return (f"{i}_{j}" for i, j in it)


from sklearn.utils import Bunch
from nilearn import plotting
from nilearn.maskers import NiftiMapsMasker, NiftiLabelsMasker
from pathlib import Path


import xml.etree.ElementTree as ET
import matplotlib.pyplot as plt
from nilearn import datasets
import pandas as pd
import nibabel as nib
import numpy as np
import os


atlas_mapping = {
    "harvard-oxford": lambda : datasets.fetch_atlas_harvard_oxford("cort-maxprob-thr25-2mm"),
    "schaefer": lambda : datasets.fetch_atlas_schaefer_2018(resolution_mm=2),
    "schaefer200": lambda : datasets.fetch_atlas_schaefer_2018(n_rois=200, resolution_mm=2),
    "schaefer100": lambda : datasets.fetch_atlas_schaefer_2018(n_rois=100, resolution_mm=2),
    "difumo": lambda : datasets.fetch_atlas_difumo(legacy_format=False),
    "smith": datasets.fetch_atlas_smith_2009,
    "msdl": datasets.fetch_atlas_msdl
}

is_soft_mapping = {
    "schaefer200": False,
    "harvard-oxford": False,
    "msdl": True,
    "GINNA": False
}

class Atlas(Bunch):
    @classmethod
    def from_kwargs(cls, name, soft, **atlas_kwargs) -> None:
        new = cls(**atlas_kwargs)

        new.is_soft = soft
        new.name = name
        return new
    
    @classmethod
    def from_name(cls, name, soft=None):
        if soft is None:
            soft = is_soft_mapping[name] 
        atlas_kwargs = atlas_mapping[name]()
        new = cls(**atlas_kwargs)
        new.is_soft = soft
        new.name = name
        new.labels_ = atlas_kwargs["labels"]
        return new

    def get_coords(self):
        if "region_coords" in self.keys():
            return self.region_coords
        elif self.is_soft:
            return plotting.find_probabilistic_atlas_cut_coords(self.maps)
        else:
            return plotting.find_parcellation_cut_coords(self.maps)
    
    def overlay(self):
        raise NotImplementedError()

    def plot(self, **plotting_kwargs):
        if self.is_soft:
            return plotting.plot_prob_atlas(self.maps, title=self.name, **plotting_kwargs)
        else:
            return plotting.plot_roi(self.maps, title=self.name, **plotting_kwargs)

    def fit_masker(self, **masker_kw):
        if self.is_soft:
            masker = NiftiMapsMasker(
                maps_img=self.maps,
                **masker_kw
            )
        else:
            masker = NiftiLabelsMasker(
                labels_img=self.maps,
                #labels=self.labels, # TODO Test that
                **masker_kw
            )
        masker.fit()
        return masker

    @property
    def labels(self):
        # The issue is that for difumo the labels
        # are in a data frame
        # Check type of labels instead?
        if self.name == "difumo":
            return self.labels_.difumo_names.to_list()
        else:
            return self.labels_

    @property
    def macro_labels(self):
        if "networks" in self.keys():
            return self.networks
        l = self.labels
        return list(map(lambda x: str(x).split("_")[2], l))


#computation of functional connectivity matrices
import numpy as np 
import pandas as pd
from scipy.spatial.distance import pdist, squareform
from scipy.stats import zscore

def compute_connectivity(data, method='pearson', matrix_form=True, to_numpy=False, **kwargs):
    """
    Compute the correlation matrix of a given dataset
    :param data: 2D numpy array
    :param method: string, 'pearson'
    :return: 2D numpy array, the correlation matrix
    :or a pandas DataFrame if nodenames are provided
    """
    nodenames = kwargs.get('nodenames', None)
    #z score data
    # data = zscore(data, axis=1)

    if method == 'pearson':
        # corr_matrix = np.corrcoef(data, rowvar=rowvar)
        corr_matrix = np.corrcoef(data, rowvar=True)
    else:
        raise ValueError('method should be pearson. The developpers were too lazy to implement another method yet.')
     
    if nodenames is not None:
        if len(nodenames) != corr_matrix.shape[0]:
            raise ValueError('Number of nodenames should be equal to the number of nodes in the data')
        return pd.DataFrame(corr_matrix, index=nodenames, columns=nodenames)
    # elif to_numpy:
    #     return corr_matrix.reshape(1,-1)
    else:
        if matrix_form == False:
            # get upper triangle
            corr_matrix = corr_matrix[np.triu_indices(corr_matrix.shape[1], k=1)]
            corr_matrix = corr_matrix.reshape(1,-1).squeeze()
        if to_numpy:
            return corr_matrix
        return pd.DataFrame(corr_matrix, columns=[i for i in range(corr_matrix.shape[1])])


from itertools import product
import numpy as np
import numpy.linalg as npl
import matplotlib.pyplot as plt
from nilearn.connectome import cov_to_corr
import matplotlib.patches as mpatches




def generate_topology(target_label, labels):
    n_regions = len(labels)
    labels = np.array(labels)

    mat = np.zeros((n_regions, n_regions))
    network_idx = np.where(labels == target_label)[0]
    mat_coords = tuple(product(network_idx, network_idx))
    for coords in mat_coords:
        mat[coords] = 1
    return mat
    

def generate_topology_net_interaction(targets, labels):
    if targets[0] == targets[1]:
        return generate_topology(targets[0], labels)

    n_regions = len(labels)
    labels = np.array(labels)
    mat = np.zeros((n_regions, n_regions))
    target_idx = []
    for target in targets:
        target_idx.append(
            np.where(labels == target)[0]
        )
    mat_coords = tuple(product(*target_idx))
    for coords in mat_coords:
        mat[coords] = 0.5
        mat[coords[::-1]] = 0.5 # Topology must be symmetric

    return mat

rng = np.random.default_rng()


def generate_correlations(n_subjects, topology, loc, scale, snr=2, from_prec=False, debug_display=False):
    correlations = []
    mask = topology > 0
    n_regions = len(topology)
    if debug_display:
        plt.hist(rng.normal(loc=loc, scale=scale, size=n_regions * 10))
        plt.title(f"Source normal distribution N({loc}, {scale})")
        plt.show()

    for _ in range(n_subjects):
        synthetic = topology.copy()
        synthetic[mask] = rng.normal(loc=loc, scale=scale, size=(mask.sum())) # TODO More realistic distribution?
        synthetic += rng.normal(loc=0, scale=np.sqrt(loc**2/snr), size=synthetic.shape) # Assuming SNR = (mu ^ 2) / (sigma ^ 2)

        synthetic += np.eye(n_regions)
        synthetic = synthetic.T @ synthetic

        np.testing.assert_almost_equal(synthetic, synthetic.T)
        eigenvalues = np.linalg.eigvalsh(synthetic)
        if eigenvalues.min() < 0:
            raise ValueError(
                "Failed generating a positive definite precision "
                "matrix. Decreasing n_features can help solving "
                "this problem."
            )

        if from_prec:
            covariance = npl.inv(synthetic)
        else:
            covariance = synthetic
            
        correlation = cov_to_corr(covariance)

        if debug_display:
            plt.hist(covariance[mask], histtype="step", color="green")
            plt.hist(correlation[mask], histtype="step", color="red")

        correlations.append(correlation) # We could yield but what's the point
    if debug_display:
        plt.title("Random values in topology mask")
        green_patch = mpatches.Patch(color='green', label='Covariance')
        red_patch = mpatches.Patch(color='red', label='Correlation')
        plt.legend(handles=[green_patch, red_patch])
        plt.show()
    return correlations


#predictors for continuous behavioral data
from sklearn.kernel_ridge import KernelRidge
from abc import ABC


class RegressionPredictor(ABC):
    # abstract methods to be implemented by subclasses
    def fit(self, X, y):
        pass

    def predict(self, X):
        pass


class KernelRidgePredictor(RegressionPredictor):
    def __init__(self, alpha=1.0, kernel='linear'):
        self.alpha = alpha
        self.kernel = kernel
        self.model = KernelRidge(alpha=self.alpha, kernel=self.kernel)

    def fit(self, X, y):
        self.model.fit(X, y)

    def predict(self, X):
        return self.model.predict(X)



from matplotlib import pyplot as plt
from matplotlib.ticker import FixedLocator
import numpy as np
import seaborn as sns


def plot_matrix(
    mat, atlas, macro_labels=True, bounds=None, cmap="seismic", axes=None
):
    """Simplified version of the plot_matrices function. Only displays
    a single matrix.

    Args:
        mat (_type_): _description_
        atlas (Bunch): sklearn bunch containing labels and
        macro labels id macro_labels is True
        macro_labels (bool, optional): _description_. Defaults to True.
        bounds (_type_, optional): _description_. Defaults to None.
        cmap (str, optional): _description_. Defaults to "seismic".

    """
    mat = mat.copy()
    n_regions = mat.shape[0]
    mat[list(range(n_regions)), list(range(n_regions))] = 0

    # In general we want a colormap that is symmetric around 0
    span = max(abs(mat.min()), abs(mat.max()))
    if bounds is None:
        bounds = (-span, span)

    if axes is None:
        fig, axes = plt.subplots(1, 1, figsize=(8, 6))

    if macro_labels:
        networks = np.array(atlas.macro_labels)

        sort_index = np.argsort(networks)
        ticks = []
        lbls = []
        prev_label = None
        for i, label in enumerate(networks[sort_index]):
            if label != prev_label:
                ticks.append(i)
                lbls.append(label)
                prev_label = label
                axes.hlines(i, 0, n_regions, colors="black", linestyles="dotted")
                axes.vlines(i, 0, n_regions, colors="black", linestyles="dotted")

        ticks.append(i + 1)

    else:
        sort_index = np.arange(n_regions)

    sns.heatmap(
        mat[np.ix_(sort_index, sort_index)],
        ax=axes,
        vmin=bounds[0],
        vmax=bounds[1],
        cmap=cmap
    )

    if macro_labels:
        axes.yaxis.set_minor_locator(FixedLocator(ticks))
        axes.yaxis.set_major_locator(FixedLocator([(t0 + t1) / 2 for t0, t1 in zip(ticks[:-1], ticks[1:])]))
        axes.xaxis.set_major_locator(FixedLocator([(t0 + t1) / 2 for t0, t1 in zip(ticks[:-1], ticks[1:])]))
        axes.set_yticklabels(lbls, rotation=0)
        axes.set_xticklabels(lbls, rotation=30)

    return axes


from nilearn.maskers import NiftiLabelsMasker

def parcellate(img, atlas):
    #img and atlas can either be a path to a nifti file or a nifti image object

    masker = NiftiLabelsMasker(labels_img=atlas, strategy='mean')

    avg_data = masker.fit_transform(img).T
    
    return avg_data


import matplotlib.pyplot as plt
import numpy as np

from itertools import combinations
import math


def reshape_pvalues(pvalues):
    l = len(pvalues)
    
    # Mat size is the positive root of :
    # n**2 - n - 2l = 0 
    # Where l is the length of pvalues array
    # and n is the square matrix size
    n = (1 + math.sqrt(1 + 8 * l)) / 2
    if n != int(n):
        raise ValueError(f"Array of lenght {l} cannot be reshaped as a square matrix")
    n = int(n)
    
    arr = np.zeros((n, n))
    pointer = 0
    for i in range(n):
        if i + pointer > pointer:
            arr[i, :i] = pvalues[pointer:pointer+i]
        pointer += i

    return arr + arr.T

def fast_hist(matrix:np.ndarray):
    """Plot values of arrays containing
    individuals correlations

    Args:
        matrix (np.ndarray): (n_subjects, n_regions, r_regions)

    """
    n_regions = matrix.shape[1]
    n_subjects = matrix.shape[0]
    fig, ax = plt.subplots(1, 1)

    # Passing the array is slower
    for i in range(n_subjects):
        tst = matrix[i, :, :].reshape((n_regions ** 2))
        ax.hist(tst, histtype="step")
    ax.set_xlim(-1, 1)
    return fig, ax





from pycirclize import Circos
from pandas import DataFrame
import math
from typing import Iterable

def make_links(matrix_df: DataFrame, zoom: int=10) -> tuple:
    links = []

    for source, row in matrix_df.iterrows():
        for target_rsn, val in row.items():
            if math.isclose(val, 0) or math.isnan(val):
                continue
            elif val > 0:
                colour = "red"
            elif val < 0:
                colour = "blue"

            offset = abs((zoom * val))
            assert offset < 10, f"{source} -> {target_rsn}, Offset = {offset}"
            print(source, target_rsn, val, offset)
            if source != target_rsn:
                links.append(
                    (
                        (source, 5 - offset / 2, 5 + offset / 2),
                        (target_rsn, 5 - offset / 2, 5 + offset / 2),
                        colour
                    )
                )
            else:
                links.append(
                    (
                        (source, 0, offset),
                        (target_rsn, 10 - offset, 10),
                        colour
                    )
                )
    return links
    

def create_chord(
    networks: Iterable,
    links: Iterable,
    **sector_kws
):
    sectors = {n: 10 for n in networks}
    circos = Circos(sectors, space=1)
    
    r_lim=(90, 100)
    for sector in circos.sectors:
        # Plot label, outer track axis & xticks
        sector.text(sector.name, **sector_kws)
        outer_track = sector.add_track(r_lim)
        color = "gray"
        outer_track.axis(fc=color, alpha=0.5)
    
    
    # Plot links
    for link in links:
        circos.link(*link[:2], color=link[2])
    
    
    fig = circos.plotfig()
    return fig


import pandas as pd
import numpy as np
from nilearn.maskers import NiftiLabelsMasker
from sklearn.metrics import adjusted_mutual_info_score, mutual_info_score
import os
import sqlite3
import itertools
import sys
sys.path.insert(1, '/homes_unix/agillig/neuroginius/neuroginius')
import retrieve_data as rtr
import functional_connectivity as FC
from tqdm.auto import tqdm


from SQLInterface import connectivity as SQLConn


def decoding_pipeline(job, **kwargs):
    debug = kwargs.get('debug', False)
    index = int(job) - 1

    atlas_name = 'schaefer200' 
    #params
    parallel = 'slurm' #whether to use slurm for massive parallelisation. If 'multiproc', will use multiprocessing?
    base_dir = '/homes_unix/agillig/Projects/DynaPred'
    ParcellatedData_db = base_dir + f'/processing/databases/SHARE_parcellated_{atlas_name}.db'
    FCData_db = base_dir + f'/processing/databases/SHARE_FC_{atlas_name}.db'

    rsDataList_file = base_dir + '/processing/MRiShare_gsreg_list.txt'
    rsDataList = np.loadtxt(rsDataList_file, dtype=str)
    SubList = [rsfile.split('subject_id_')[1].split('/')[0] for rsfile in rsDataList]
    SubList.sort()
    Atlases_dir = '/homes_unix/agillig/Atlases' ##TO BE UPDATED
    # atlas_file = '/homes_unix/agillig/Atlases/M5_clean2_ws/RSN_N41_atlas_M5_clean2_wscol.nii'
    atlas_file = Atlases_dir + '/Schaefer2018/Schaefer2018_200Parcels_7Networks_order_FSLMNI152_2mm.nii.gz'
    #INIT DIRS

    #INIT VARS 

    



    ## OPTIONAL: specify type of decoding to be performed via indexing array

    ##


    ###STEP 1: DATA PREPARATION
    #temporary; this is just for me
    # # find the rsDataList files that correspond to the SubID
    # ID_list = [rsfile.split('subject_id_')[1].split('/')[0] for rsfile in rsDataList]
    # #filter = np.isin(filter, SubID)
    # SubID = ID_list[index]
    # rsData_file = rsDataList[index]
    # rsData_file = ''.join(rsData_file)
	
    # if debug:
    #     rsData_file = '/homes_unix/agillig/Projects/DynaPred/data/tmp/SHARE001.nii.gz'

    # print(f'data file: {rsData_file}')



    #TODO: add interface that #1 check if there are missing subjects according to the provided list,
    #2 computes the missing ones, 3 retrieves everything

    # check if all parcellations are here
    missing = rtr.check_tables_in_sql(SubList, ParcellatedData_db)
    print(f'{len(missing)} missing parcellations')
    #temporary too, TODO: implement in neuroginius
    # will work only on pipeau, obviously. but should work
    # if len(missing) > 0:
    #     print(f'{len(missing)} missing parcellations')
    #     for subject in missing:
    #         file = rsDataList[SubList.index(subject)]
    #         print(subject, file)
    #         print(f'computing parcellation for {subject}')
    #         rtr.retrieve_parcellated_data(subject,
    #                                       ParcellatedData_db,
    #                                       DataFile=file,
    #                                       atlas=atlas_file)

    # rsData = rtr.retrieve_all_parcellated_data(ParcellatedData_db,
    #                                            SubList, 
    #                                            compute_missing=False)

    ## TO BUILD: frame filtering thanks to edge time series / RSS

    # print('computing functional connectivity')
    # for data in tqdm(rsData):
    #     tmp_fc_data = FC.compute_connectivity(data, method='pearson', matrix_form=False)

    # print('loading functional connectivity matrices')
    SubList = [sub for sub in SubList if sub not in missing]
    SubList = SubList[:100]
    # FCData = rtr.get_fc_data(FCData_db, SubList, check_missing=True, compute_missing=True, parcellation_db=ParcellatedData_db, matrix_form=False, atlas=atlas_file)

    FCData, SubList_new = SQLConn.ConnectivityFromParcellationSQL(ParcellatedData_db,
                                                                  SubList=SubList, 
                                                                  method='pearson', 
                                                                  n_jobs=16,
                                                                  ConnectivityDb=FCData_db
                                                                  )
    print(type(FCData))
    print(len(FCData))
    print(SubList_new[0], SubList[0], FCData[0])

    differences = [(i, a, b) for i, (a, b) in enumerate(zip(SubList, SubList_new)) if a != b]

    for i, a, b in differences:
        print(f"Element at index {i} is not the same: {a} in SubList, {b} in SubList_new")
    ## TO BUILD: decoding with nested cv & non parametric testing

    ## To build: permutation feature importance to interpret decoding drivers

    ## end goal: "multivariate" optimization to find the network that best predicts individual differences


    print('end of file, done')


if __name__ == '__main__':
    decoding_pipeline(1)


from joblib import Parallel, delayed
import sqlite3
from SQLInterface import RetrieveData
import functional_connectivity as FC
from tqdm.auto import tqdm
from concurrent.futures import ThreadPoolExecutor
import pandas as pd
import warnings


 

def ConnectivityToSQL(FCData_db, SubList, if_exists='continue', n_jobs=None):
    """
    Compute the functional connectivity matrices for a list of subjects and save them to a SQL database
    :param FCData_db: string, path to the database where the functional connectivity matrices will be saved
    :param SubList: list of strings, the list of subjects
    :param if_exists: string, 'continue' or 'replace'
    """
    pass

def ConnectivityFromParcellationSQL(ParcellationDb, SubList=None, method='pearson', n_jobs=1, ConnectivityDb=None, in_memory=False):
    """
    Compute the functional connectivity matrices from a parcellation database
    :param ParcellationDb: string, path to the database where the parcellation data is stored
    :param SubList: list of strings, the list of subjects
    :return: 3D numpy array, the functional connectivity matrices
    """
    if type(SubList) == 'str':
        SubList = [SubList]

    print('loading parcellated data')
    DataList, SubIDs = RetrieveData.RetrieveTablesFromSQL(ParcellationDb, SubList, index_col='ROI', n_jobs=n_jobs)
    #!!!!! fix the orientation of parcellated data directly in df
    warnings.warn('Incorrect orientation of parcellated data, fix the saving parcellation step')
    DataList = [data.T for data in DataList]

    print(DataList[0].shape)
    print('computing functional connectivity')
    FCData = Parallel(n_jobs=n_jobs)(delayed(FC.compute_connectivity)(data, method=method, matrix_form=False) for data in tqdm(DataList))

    FCData = pd.concat(FCData, axis=0).T
    FCData.columns = [i for i in range(FCData.shape[1])]
    print(FCData.columns)
    print(FCData.shape)
    if ConnectivityDb is not None:
        print('saving to database')
        # [RetrieveData.SaveToSQLTable(data, ConnectivityDb, index_label='ID') for data, ID in tqdm(zip(FCData, SubIDs))]
        
        if not in_memory:
            # Parallel(n_jobs=1)(delayed(RetrieveData.SaveToSQLTable)(data, ConnectivityDb, index_label='ID') for data, ID in tqdm(zip(FCData, SubIDs)))
            RetrieveData.SaveToSQLTable(FCData, ConnectivityDb, index_label='ID')
        else:
            conn = sqlite3.connect(':memory:', check_same_thread=False)
            # Parallel(n_jobs=n_jobs)(delayed(RetrieveData.save_to_sql_in_memory)(data, conn, index_label='ID') for data, ID in tqdm(zip(FCData, SubIDs)))
            # Use the function in your parallel job
            with ThreadPoolExecutor(max_workers=n_jobs) as executor:
                futures = [executor.submit(RetrieveData.save_to_sql_in_memory, data, conn, index_label='ID') for data, ID in tqdm(zip(FCData, SubIDs))]
                results = [future.result() for future in futures]
            #dump to file
            conn_disk = sqlite3.connect(ConnectivityDb)
            conn.backup(conn_disk)
        print(f'saved to {ConnectivityDb}')
    return FCData, SubIDs



import sqlite3
import pandas as pd
import parcellate as par
from pathlib import Path


def retrieve_parcellated_data(SubID, db_file, **kwargs):
    """
    Retrieve parcellated data from the database
    """
    atlas = kwargs.get('atlas', None)

    # get the data from the database
    query = f"SELECT * FROM parcellated_data WHERE SubID = '{SubID}'"

    conn = sqlite3.connect(db_file)
    # Execute a SQL query and load the result into a DataFrame
    data = pd.read_sql_query(query, conn)

    if sequences_df.shape[0] != 0:
        return data

    print(f'{SubID} not in the database, computing and saving parcellation')
    # if the data is not in the database, parcellate the data
    if atlas is None:
        raise ValueError(f'an atlas is required to parcellate the data. subject {SubID} not found in the database and no atlas provided')
    atlas = Path(".").absolute().parent / "atlases/"

    data = par.parcellate(SubID, atlas)
    data = pd.DataFrame(data, index=[SubID])
    data.to_sql('parcellated_data', conn, if_exists='append', index=True, index_label='SubID')
    
    conn.commit()
    conn.close()

    return data



def compute_network_freqs(job, **kwargs):
    debug = kwargs.get('debug', False)
    index = int(job) - 1 
    import pandas as pd
    import numpy as np
    from nilearn.maskers import NiftiLabelsMasker
    from sklearn.metrics import adjusted_mutual_info_score
    import os
    import sqlite3

    base_dir = '/homes_unix/agillig/Projects/DynaPred'
    #SubList_file = base_dir + '/processing/SHARE_tmp_SubID.csv'
    #SubList = pd.read_csv(SubList_file, header=0, index_col=0).to_numpy().flatten()

    #SubID = SubList[index]

    rsDataList_file = base_dir + '/processing/MRiShare_gsreg_list.txt'
    rsDataList = np.loadtxt(rsDataList_file, dtype=str)

    # find the rsDataList files that correspond to the SubID
    ID_list = [rsfile.split('subject_id_')[1].split('/')[0] for rsfile in rsDataList]
    #filter = np.isin(filter, SubID)
    SubID = ID_list[index]
    rsData_file = rsDataList[index]
    rsData_file = ''.join(rsData_file)
	
    if debug:
        rsData_file = '/homes_unix/agillig/Projects/DynaPred/data/tmp/SHARE001.nii.gz'

    print(f'data file: {rsData_file}')

    # extract data from the file using NiftiLabelMasker (mean within parcels)
    atlas_file = '/homes_unix/agillig/Atlases/Schaefer2018/Schaefer2018_200Parcels_7Networks_order_FSLMNI152_2mm.nii.gz'

    LabelMasker = NiftiLabelsMasker(labels_img=atlas_file, strategy='mean',
                  standardize='zscore_sample')

    print(f'loading data for subject {SubID}')
    rsData = LabelMasker.fit_transform(rsData_file).T
    print('done')


    ## create network partitions ##
    print('creating network partitions')
    Atlases_dir = '/homes_unix/agillig/Atlases' ##TO BE UPDATED
    centroids_file = Atlases_dir + '/Schaefer2018/Schaefer2018_200Parcels_7Networks_order_FSLMNI152_2mm.Centroid_RAS.csv' 

    centroids = pd.read_csv(centroids_file)

    networks = centroids['ROI Name'].to_numpy()
    networks = [netstr.split('_')[2] for netstr in networks]
    networks = np.array(networks)

    networks_names = np.unique(networks)

    templates = []
    for network in networks_names:
        print(f'creating partition for network {network}')
        template_mask = np.array(networks == network).astype(int)

        pairwise_products = np.outer(template_mask, template_mask)

        indices = np.triu_indices(pairwise_products[0].shape[0], 1)
        template_edges = np.array(pairwise_products[indices])

        templates.append(template_edges)

    templates = pd.DataFrame(np.array(templates).T, columns=networks_names)


    # compute the edge time series & bipartition
    print('computing edge time series')
    ### step1 : pairwise products between the z scores of the time series - > edge time series
    pairwise_products = np.array([np.outer(rsData[:, i], rsData[:, i]) for i in range(rsData.shape[1])])

    indices_uppertriangle = [np.triu_indices(pairwise_products[0].shape[0], 1) for i in range(pairwise_products.shape[0])]
    edge_ts = np.array([prod[ind] for prod, ind in zip(pairwise_products, indices_uppertriangle)])

    bipartitions = (edge_ts > 0).astype(int)


    # compute the mutual information at each time point between the edge time series and the bipartition
    print(f'computing mutual information with network templates')
    mutual_info = []
    for i, network in enumerate(np.unique(networks)):
        template = templates[network]

        if debug:
            print(f'shape of template: {template.shape}, shape of bipartitions: {bipartitions.shape}')
        tmp_mutual_info = [adjusted_mutual_info_score(x, template) for x in bipartitions]
        tmp_mutual_info = np.array(tmp_mutual_info)
        mutual_info.append(tmp_mutual_info)

    # threshold of significance of the mutual information; computed using random permutations of the network labels
    value_significance_file = base_dir + '/processing/SHARE_tmp_value_significance.csv'
    value_significance = np.loadtxt(value_significance_file)

    print(f'computing frequency of networks occurences')
    freqs_occurences = []
    for mi in mutual_info:
        is_significant = (mi > value_significance).astype(int)
        n_occurences = len(mi[is_significant == 1])
        freq_occurence = n_occurences / len(is_significant)
        freqs_occurences.append(freq_occurence)

    freqs_occurences = np.array(freqs_occurences).reshape(1, -1)
    if debug:
        print(f'freqs_occurences shape: {freqs_occurences.shape}')
    freqs_occurences = pd.DataFrame(freqs_occurences, index=[SubID], columns=networks_names)


    print(freqs_occurences)
    # find a way to save: databasing? simple file?
    #<<use sqlite databasing

    db_file = base_dir + '/processing/network_freqs/database/SHARE_network_freqs.db'

    os.makedirs(os.path.dirname(db_file), exist_ok=True)

    conn = sqlite3.connect(db_file)

    # Append the DataFrame to the table in the SQLite database
    freqs_occurences.to_sql('network_freqs', conn, if_exists='append', index=True, index_label='SHARE_ID')

    conn.commit()
    conn.close()



from itertools import product, combinations
import numpy as np
import itertools
from typing import Iterable, Tuple

import pandas as pd

from neuroginius.atlas import Atlas
from neuroginius.iterables import unique, all_equal, join, all_connectivities

def edge_format(block, labels):
    unstacked = block.unstack().reset_index().rename(columns={0: "FC"})
    unstacked["edge"] = unstacked["node_a"] + "_" + unstacked["node_b"]

    target_cols = tuple(join(all_connectivities(labels)))
    triangle_mask = unstacked.edge.isin(target_cols)
    c = unstacked[triangle_mask].loc[:, ["edge", "FC"]]
    return c.set_index("edge").T

def network_to_network_connectivity(matrix, network_to_idx, pairing_func=combinations):
    """

    Args:
        matrix (_type_): Matrix should have the block structure 
        described in network_to_idx!
        network_to_idx (_type_): _description_

    Yields:
        _type_: _description_
    """
    for network_a, network_b in pairing_func(network_to_idx.index, 2):
        loc_a, loc_b = network_to_idx[network_a], network_to_idx[network_b]
        connectivity = matrix[loc_a[0]:loc_a[1], loc_b[0]:loc_b[1]].mean()
        yield network_a, network_b, connectivity

def edge_counts(block):
    n_positive_edges = np.count_nonzero(block > 0)
    n_negative_edges = np.count_nonzero(block < 0)
    block_activation = (n_negative_edges > 0) or (n_positive_edges > 0)
    return n_positive_edges, n_negative_edges, block_activation
    

def block_block(matrix, network_to_idx, aggregating_func=edge_counts):
    for network_a, network_b in product(network_to_idx.index, network_to_idx.index):
        loc_a, loc_b = network_to_idx[network_a], network_to_idx[network_b]
        block = matrix[loc_a[0]:loc_a[1], loc_b[0]:loc_b[1]]

        yield network_a, network_b, *aggregating_func(block)


def macro_matrix(matrix, network_to_idx):
    gen = block_block(matrix, network_to_idx, aggregating_func=lambda block : (block.mean(),))
    comparisons = pd.DataFrame(gen, columns=["node_a", "node_b", "connectivity"])
    pivoted = comparisons.pivot(index="node_a", columns="node_b")
    return pivoted.loc[:, "connectivity"]

def groupby_blocks(matrix, atlas):
    ticks, sort_index = group_by_networks(atlas.macro_labels)
    matrix_sort = np.ix_(sort_index, sort_index)
    sorted_matrix = matrix[matrix_sort]
    new_labels = sorted(tuple(unique(atlas.macro_labels)))

    network_to_idx = pd.Series(dict(zip(
        new_labels,
        itertools.pairwise(ticks)
    )))
    return macro_matrix(sorted_matrix, network_to_idx), new_labels

def group_by_networks(macro_labels):
    networks = np.array(macro_labels)
    sort_index = np.argsort(networks)

    ticks = []
    lbls = []
    prev_label = None
    for i, label in enumerate(networks[sort_index]):
        if label != prev_label:
            ticks.append(i)
            lbls.append(label)
            prev_label = label

    ticks.append(i+1)
    return ticks, sort_index

def group_groupby(matrices: Iterable, atlas: Atlas) -> Tuple[np.ndarray, list[str]]:
    """Apply grouping by networks to an iterable of connectivity matrices

    Returns:
        Tuple[np.ndarray, list[str]]: Array of new connectivity matrices with
        the corresponding macro labels
    """
    res = (groupby_blocks(mat, atlas) for mat in matrices)
    blocks, labels = zip(*res)
    
    assert all_equal(labels)
    labels = labels[0]

    return blocks, labels


from matplotlib import pyplot as plt
import numpy as np

def plot_fdr(pvalues, pvalues_corr, alpha=0.05):
    fig, ax = plt.subplots()
    x = np.linspace(0, alpha, 100)
    ax.plot(x, x, color="black", label="Reference")
    ax.plot(np.sort(pvalues), np.sort(pvalues_corr), label="Corrected")
    ax.set_xlim(0, alpha)
    ax.hlines(
        y=alpha,
        xmin=0,
        xmax=alpha,
        colors="red", 
        label=f"Threshold $\\alpha = {alpha}$",
        linestyles="--"
    )
    ax.set_xlabel("pvalues_raw")
    ax.set_ylabel("pvalues_corrected")
    ax.set_title(f"False discovery correction")
    ax.legend()
    return fig